#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <unistd.h>
#include "config.h"

/*
Esercizio2
	un programmino di esempio sull'uso delle pipe e le configurazioni
	client/server in C

server_pipe.c
	server reale, cicla prendendo input da una pipe (l'input comprende
	un comando e il PID del client) e, se il comando e` buono,
	eseguiendo tramite exec il comando stesso; l'output del comando
	e` inviato al client, tramite una pipe il cui nome e` ricavato dal
	PID precedentemente passato

Copyright (C) 1998 Marco Gaiarin (gaio@pluto.linux.it)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.		*/


// prende in buf una stringa nel formato:
// 	PID # comando #		(spazi ad libitum)
// e *APPENDE* in cpname e cmd PID e comando rispettivamente
// 
// ritorna il numero di cancelletti trovati, quindi se ritorna 2 il comando
// e` presumibilmente buono
int parsebuf (const char *buf, char *cpname, char *cmd) {

int	i = 0,	// indice sull array of char
	d = 0,	// numero di delimitatori
		// evito di passare un parametro in piu`... ;)))
	bufl = strlen (buf);

// ciclo finche` c'e` roba oppure trovo il secondo #...
while ( (i < bufl) && (d < 2) ) {

#ifdef DEBUG
	printf ("server: carattere %c, numero delimitatori %u\n",buf[i],d);
#endif

switch (buf[i]) {
case	'#':	// delimitatore
	
	d++;

	break;

case	' ':	// compatto gli spazi

	break;

default:	// vedo dove devo mettere...

	if ( d == 0 ) // e` il PID
	  if ( (buf[i] >= '0') && (buf[i] <= '9') )
	    strncat(cpname, &buf[i], 1);
	  else {
	    perror("server_pipe: client usa PID non numerico!");
	    exit(SERR);
	  }
	else if ( d == 1) 	// e` il comando... sono di bocca buona tanto
				// c'e` il parser...
	  strncat(cmd, &buf[i], 1);

	break;

} // switch

i++;

} // while

return (d);

} // parsebuf



// stampa un po' di help se il parser dice che il comando non e` buono
void help(void) {

int i;

printf ("server_pipe comes with ABSOLUTELY NO WARRANTY;\n");
printf ("This is free software, and you are welcome to redistribute it\n");
printf ("under certain conditions; see GNU GPL version 2 or above for detail\n\n");
printf ("mi dispiace, ma il comando non e` valido o perlomeno io\n");
printf ("non sono cosi` stolto fare exec() alla cieca... ;)\n\n");
printf ("i comandi validi sono:\n");
for ( i=1; i <= PCMD; i++ ) {
printf ("\t%s",strrchr(getcmd(i),'/')+1);
if ( ((i-1) % 5) == 4 )
	printf("\n");
}
printf("\n");

} // help()



// questo e` il server vero e proprio.
// cicla indefinitivamente aprendo e chiudendo la master (server) pipe dove
// riceve i comandi, ed eseguendo il comando se questo e` sintatticamente
// corretto ed inoltre e` buono
// i comandi buoni sono quelli definiti in config.h
void server_pipe(void) {

int	pfd,		// server pipe file descriptor
	cmdn,		// numero del comando
	i;
char	cpname[PNLEN],	// client's pipe name
	buf[BUFLMAX],	// buffer per il comando
	cmd[CLEN];	// comando da eseguire

#ifdef DEBUG
	int	bufl;	// lunghezza di quanto letto, per debug
#endif


// il server, come tutti i server, non fa altro che fare un bel loop
// aspettando comandi da eseguire
while ( 18 == 18 ) {	// mio giorno di nascita ;)

// Ora apriamo la pipe; qui il programma si blocca in attesa di un
// input; questa pipe e` aperta in sola lettura, questa deve solo
// ricevere i PID dei clients che cercano di collegarsi e il comando
if ( (pfd = open(PNAME, O_RDONLY)) < 0 ) {
	perror("server_pipe: errore aprendo il pipe");
	exit(OCERR);
}
#ifdef DEBUG
	printf("server: master pipe aperta, FD=%u\n",pfd);
#endif

// ora leggiamo PID e comando.
// leggo un paccozzo di roba
#ifdef DEBUG
	bufl = read(pfd, &buf, BUFLMAX);
	printf("server: ho letto %s di %u caratteri\n",buf,bufl);
#else
	read(pfd, &buf, BUFLMAX);
#endif

// inizalizzo le variabili che mi servono...
sprintf(cpname, "%s%s", PNAME, ".");
cmd[0] = 0;

// ogni tanto, anche se e` tutto chiuso e ho letto tutto, si ostina
// a fare qualche giro leggendo niente, o leggendo schifezze
// se ci sono stati due delimitatori, il comando e`molto probabile
// che sia buono veramente e solo cosi` viene processato
if ( parsebuf(buf, cpname, cmd) == 2 ) {

#ifdef DEBUG
	printf("server: ricevuto pid, uso pipe: %s \n",cpname);
	printf("server: ricevuto comando: %s\n",cmd);
#endif

// ok, abbiamo tutto quello che ci serve, se il client ha scritto ha
// gia` creato la pipe e quindi ci basta fare un fork al cui interno
// gestiamo le comunicazioni.
switch ( fork() ) {

case 	-1:	// errore!

	perror("server_pipe: errore facendo fork");
	close(pfd);
	exit(FWERR);

case	0:	// figlio!

	// parso il comando
	cmdn = parse(&cmd);
	if (cmdn == 0)
		printf("server_pipe: ricevuto comando numero %u, eseguo help()\n",cmdn);
	else
		printf("server_pipe: ricevuto comando numero %u, eseguo %s\n",cmdn,getcmd(cmdn));
#ifdef DEBUG
	printf("server: ho ricevuto il comando %s che ha numero %u\n", cmd,cmdn);
	printf("\te quindi comando %s e path %s\n",cmds[cmdn-1],cmdpath[cmdn-1]);
#endif

	// apriamo in sola scrittura la pipe del client
	// e assegnamola a stdout, notare il controllo stretto
	// da qui in poi niente printf di debug, altrimenti finiscono
	// sulla pipe...
	close(STDOUT);
	if ( open(cpname, O_WRONLY) != STDOUT ) {
		perror("server_pipe: errore aprendo il pipe del client");
		close(pfd);
		exit(RWERR);
	}
	close(STDIN);

	// ok, se sono qui in qualche maniera il figlio si e` sincronizzato,
	// quindi chiudo la server_pipe
	// cosi` chiudo la pipe del figlio, ma tanto fa...
	close(pfd);

	// azione vera e propria...
	switch ( cmdn ) {
	case	0:	// nessun comando valido:

		help();
		
		// qui esce il figlio!!!
		exit (0);

	default:	// comando valido!!!

#ifdef DEBUG
		printf ("execl(%s, %s, NULL);\n",getcmd(cmdn),getcmd(cmdn));
#endif
		execl(getcmd(cmdn),getcmd(cmdn),NULL);

		// chiaramente, per ogni exec che si rispetti... ;)))
		perror("server_pipe: errore exec, non dovrei essere qui");
		exit(OERR);	

	} // switch per il parser

default:	// il padre:
	
	// recupero un po' di zombies...
	// questo perche` un figlio diventa zombie e ci resta finche`
	// suo padre non va a vedere come e` morto (con wait*).
	// WNOHANG semplicemente lo fa in maniera nonblocking, dando
	// il colpo di grazia a tutti
	while ( (waitpid(-1, &i, WNOHANG) != 0) );

	// for sanity
	sleep(1);
	
	// altrimenti quella del padre chi la chiude?!
	close(pfd);

} // switch del fork()

} // if condizione di bonta` sul buffer

} // while (18 == 18)

// chiaramente, per ogni loop infinito che si rispetti... ;)))
perror("server_pipe: errore loop, non dovrei essere qui");
exit(OERR);

} /* end server_pipe() */
